/*
	ProWesS, PROGS Windowing System

	ProWesS multi file viewer test program

	(c) PROGS, Professional & Graphical Software
	written by Joachim Van der Auwera
	started 7 September 1995
	26.10.1995 The "<<" and ">>" items are now put inside the outline
*/

#include "ProWesS_h"

typedef struct _List {
	struct _List *next;
	struct _List *prev;
	char name[IO_MAXFULLNAME];
	} List;
#define BLOCKFACTOR 100

typedef struct {
	PWObject files, it, outl;
	char *base; int length;
	Chunks chunks;
	List *list;
	List *current;
	char dir[IO_MAXDIRECTORY];
	} Global;

/* forward declarations */
Error buildlist(PWObject object, char *dir, PWObject menu);
Error previous(PWObject object);
Error next(PWObject object);
Error display(Global *g);

Error init()
{
	PWObject obj;
	PWObject box;
	Global g;
	Error err;
	List *item;

	/* initialise Global object */
	MEMClear(sizeof(Global),&g);
	err=MEMUserInit(sizeof(List),BLOCKFACTOR,&g.chunks);
	if (err) return err;
	err=MEMUserAllocate(g.chunks,&g.list);
	if (err) return err;
	g.list->next=g.list->prev=g.list;

	/* create fileselect object */
	err=PWCreate(NULL,&g.files,PW_TYPE_FILESELECT,
				PW_FILESELECT_MULTIPLE,
				PW_FILESELECT_ACTION, buildlist,
				PW_GLOBAL_AUXILIARY, &g,
				NULL);
	if (err) return err;

	/* but the user shoud be able to put the window to sleep */
	err=PWQuery(g.files,PW_FILESELECT_OUTLINE,&obj);
	if (err) return err;
	err=PWChange(obj, PW_OUTLINE_SLEEP, NULL);
	if (err) return err;

	/* create window for viewing */
	err=PWCreate(NULL,&g.outl,PW_TYPE_OUTLINE,
				PW_OUTLINE_SLEEP,
				PW_OUTLINE_QUIT,
				PW_OUTLINE_QUIT_KEYPRESS, 27,
				PW_GLOBAL_AUXILIARY, &g,
				NULL);
	if (err) return err;
	err=PWCreate(g.outl,&g.it,PW_TYPE_INFOTEXT,
				PW_INFOTEXT_WIDTH_FS, short2pt(30),
				PW_INFOTEXT_LINES, 12,
				NULL);
	if (err) return err;
	err=PWQuery(g.outl,PW_OUTLINE_BOX_LEFT,&box);
	if (err) return err;
	err=PWCreate(box,&obj,PW_TYPE_LOOSE_ITEM,
				PW_LOOSE_CHANGE_STATUS, FALSE,
				PW_LOOSE_TEXT, "<<",
				PW_LOOSE_KEYPRESS, '<',
				PW_LOOSE_ACTION_HIT, previous,
				NULL);
	if (err) return err;
	err=PWQuery(g.outl,PW_OUTLINE_BOX_RIGHT,&box);
	if (err) return err;
	err=PWCreate(box,NULL,PW_TYPE_LOOSE_ITEM,
				PW_LOOSE_CHANGE_STATUS, FALSE,
				PW_LOOSE_TEXT, ">>",
				PW_LOOSE_KEYPRESS, '>',
				PW_LOOSE_ACTION_HIT, next,
				NULL);

	while (err==ERR_OK)
	{
		/* clear list with current items */
		for (item=g.list->next ; item!=g.list ; )
		{
			item=item->next;
			MEMUserRelease(g.chunks, item->prev);
		}
		g.list->next=g.list->prev=g.list;

		/* let the user select some files */
		err=PWChange(g.files,PW_FILESELECT_ACTIVATE,NULL);
		if (err) return err;

		/* none -> quit */
		if (g.list->next==g.list)
			break;

		/* view the files */
		g.current=g.list->next;
		display(&g);
		err=PWActivate(g.outl);
	}
	return err;
}

/* now the event handlers */

Error buildlist(PWObject object, char *dir, PWObject menu)
{
	Global *g;
	char *item;
	List *new, *curr;
	Error err;

	PWQuery(object,PW_GLOBAL_AUXILIARY,&g);
	STRCopy(dir,g->dir);
	curr=g->list;

	PWQuery(menu,PW_MENU_SELECTED_FIRST,&item);
	while (item)
	{
		err=MEMUserAllocate(g->chunks,&new);
		if (err) return ERR_OK; 	/* hey, we'll have at least BLOCKFACTOR */
		STRCopy(item,new->name);
		new->next=g->list;
		new->prev=curr;
		curr->next=new;
		g->list->prev=new;
		curr=new;
		PWQuery(menu,PW_MENU_SELECTED_NEXT,&item);
	}
	return ERR_OK;
}

Error display(Global *g)
{
	Channel ch;
	Size size;
	Error err;

	err=IOOpenPath(g->current->name,OPEN_READ,g->dir,&ch);
	if (err) return err;
	IOLength(ch,&size);
	size++; 						/* room for terminating zero */
	if (size>g->length)
	{
		if (g->base)
			MEMRelease(g->base);
		g->base=NULL; g->length=0;
		err=MEMAllocate(size,&g->base);
		if (err) goto exitload;
		g->length=size;
	}
	IOFileLoad(ch,size,g->base);
	g->base[size-1]='\0';
	PWChange(g->outl,PW_OUTLINE_TITLE_TEXT,g->current->name,NULL);
exitload:
	IOClose(ch);
	PWChange(g->it,PW_INFOTEXT_TEXT,g->base,NULL);
	return err;
}

Error previous(PWObject object)
{
	Global *g;
	Error err=ERR_OK;

	PWQuery(object,PW_GLOBAL_AUXILIARY,&g);
	if (g->current->prev!=g->list)
	{
		g->current=g->current->prev;
		err=display(g);
	}
	return err;
}

Error next(PWObject object)
{
	Global *g;
	Error err=ERR_OK;

	PWQuery(object,PW_GLOBAL_AUXILIARY,&g);
	if (g->current->next!=g->list)
	{
		g->current=g->current->next;
		err=display(g);
	}
	return err;
}
