/*
	ProWesS, PROGS Windowing System

	edit line of text type for ProWesS, "EDLI"

	(c) PROGS, Professional & Graphical Software
	written by Joachim Van der Auwera
	started 16 July 1995
	26.10.1995 fixed problem with not restoring PFPrintEscape in time
	18.12.1995 now also use Draw routine from PW_TYPE_iTEM
	23.03.1996 adjusted for internal use of pixel values
	11.04.1996 use PWHandleDefine && use ProWesS defaults
	07.05.1996 keypress now cleared when done
	11.05.1996 PW_EDLINE_ACTIVATE added
	07.06.1996 PW_EDLINE_ACTION_TEST added, including three default actions
	02.07.1996 PW_EDLINE_ACTION_TEST routine now also called before first key
	25.09.1996 PW_EDLINE_PRINT_ESCAPE added
	25.11.1996 line height now better
	07.12.1996 use PFPShow/Width/FontBbox

	13.01.1997 changes to allow multiple lines (wl)
			   changes in create, draw, change, determinesize (wl)
	15.01.1997 check on 0xd8 + 0xd0 and not twice on 0xd0 (wl)
	15.01.1997 up/down can move to other edline again
	16.01.1997 used natural numbering for the line numbers in set/get_line

	04.02.1997 edlines & dedlines in one single file. Many changes
			   borders around edlines are drawn when cursor gets to them even
			   via edline_next, up etc..
			   line_get and array_get no longer change the current line (else
			   plays havoc with curpos for dedlines) (wl)
	04.03.1997 EDLINE_AFTER works again
*/

#include <mem.h>
#include <str.h>
#include <ctype.h>
#include <io.h>
#include "PWHandler.h"

#define DEDTYPE PW_TYPE_DIRECT_EDLINE
#define EDTYPE PW_TYPE_EDLINE
#define catch(X) do { if (err=(X)) return err; } while (0)

/* global statusses */

#define p100 short2pt(100)
ColourRGB ink	= { 0,    0,    0    }; char Kink  =FALSE;
ColourRGB paper = { p100, p100, p100 }; char Kpaper=FALSE;
char font[PF_MAXFONTNAME]=""; /* name of PROforma font to use */
pt fontsize=0;						/* fontsize to use */
pt baseline=0;						/* font baseline position (in pixels) */
pix height=0;						/* height of font */
short bufflen=256;					/* default maximum size of strings */
pt xsiz=short2pt(120);				/* default minimum size of edline object */
char recalculate=TRUE;				/* set when baseline and height should be determined */
int always_type=0;					/* 0 diff, 1 always ed, 2 always ded */
PWTypeDef *border;
/* loose item object definition */

typedef struct _Object {
	PWBaseObject;
	int bufflen;			/* length of one line buffer */
	short lines;			/* number of lines (0 = no lines - impossible) */
	int current_line;		/* line we are currently on, starting at 0 */
	pix xsiz;			/* minimum width of item */
	PWObject keypress;		/* keypress attached to this object */
	Error (*DoAfter)(PWObject object);	/* call after pressing <enter> */
	Error (*DoAction)(PWObject object); /* action for when DO on edline */
	Error (*ActionTest)(PWObject object, EdlineInfo *info);
	struct _Object *next, *prev, *up, *down, *after, *dummy;
	int escape; 		/* status for PFPrintEscape */
	int is_edline;			/* flag: edline or deadline */
/* the following is for dedlines only */
	int cursor; 		/* cursor position, before string[cursor] */
	short curpos;			/* cursor position, in window */
	short offset;			/* x pageorigin where string was displayed */
	int length; 		/* length of string */
/* end dedlines specific part only */
	char string[2]; 	/* string to edit - length==bufflen */
} Object;


/* common function for ded and edlines - deals with one char */
Error DealWithChar(Object *obj, EdlineInfo *info, Gstate gstate)
{
	char w[2]={0,0}; short cw;
	/* put it in the string */
	if (isprint(info->key) && info->length<obj->bufflen-2)
	{
		/* put it in the string */
		char *pos=&info->string[info->cursor];
		char *end=&info->string[++info->length];
		while (end>=pos) { *(end+1)=*end; end--; }
		*pos=info->key;
		info->cursor++;
		w[0]=info->key;
		PFPStringWidth(gstate,w,&cw);
		obj->curpos+=cw;
	}
	else switch(info->key)
	{
	case '\n':
		obj->dummy=obj->after?obj->after:obj;
		break;
	case '\t':
		obj->dummy=obj->next;
		break;
	case (char)0xfd:
		obj->dummy=obj->prev;
		break;
	case (char)0xd0:		/* up: goto previous line */
		if (obj->current_line)
		{
			obj->dummy=obj;
			obj->current_line--;
		}
		else
			obj->dummy=obj->up;
		break;
	case (char)0xd8:		/* down: goto next line */
		if(obj->current_line<obj->lines-1)
		{
			obj->dummy=obj;
			obj->current_line++;
		}
		else
			obj->dummy=obj->down;
		break;

	case (char)0xc0 : /* left */
		if (info->cursor)
		{
			info->cursor--;
			w[0]=info->string[info->cursor];
			PFPStringWidth(gstate,w,&cw);
			obj->curpos-=cw;
		}
		break;
	case (char)0xc1 : /* alt left - cursor to start of line */
		info->cursor=0;
		obj->curpos=0;
		break;
	case (char)0xc2 : /* ctrl left - backspace */
		if (info->cursor)
		{
			char *pos=&info->string[info->cursor];
			w[0]=*(pos-1);
			info->cursor--;
			PFPStringWidth(gstate,w,&cw);
			obj->curpos-=cw;
			while (*pos) { *(pos-1)=*pos; pos++; }
			*(pos-1)='\0';
			info->length--;
		}
		break;
	case (char)0xc3 : /* ctrl alt left - delete start of line */
		{
			char *pos=info->string;
			int i=info->length-info->cursor;
			for(;i;i--) *pos++=*(pos+info->cursor);
			*pos=0;
			info->length-=info->cursor;
			info->cursor=0;
			obj->curpos=0;
		}
		break;
	case (char)0xc4 : /* shift left */
		{
			int gotspace=info->string[info->cursor-1]==' ';
			while ( info->cursor &&
					( (info->string[info->cursor-1]==' ' &&  gotspace) ||
					  (info->string[info->cursor-1]!=' ' && !gotspace) ) )
			{
				info->cursor--;
				w[0]=info->string[info->cursor];
				PFPStringWidth(gstate,w,&cw);
				obj->curpos-=cw;
				if (gotspace) gotspace=info->string[info->cursor-1]==' ';
			}
		}
		break;
	case (char)0xc6 : /* shift ctrl left */
		{
			int oldcursor=info->cursor;
			int gotspace=info->string[info->cursor-1]==' ';
			while ( info->cursor &&
					( (info->string[info->cursor-1]==' ' &&  gotspace) ||
					  (info->string[info->cursor-1]!=' ' && !gotspace) ) )
			{
				info->cursor--;
				w[0]=info->string[info->cursor];
				PFPStringWidth(gstate,w,&cw);
				obj->curpos-=cw;
				if (gotspace) gotspace=info->string[info->cursor-1]==' ';
			}
			if (info->cursor!=oldcursor)
			{
				char *pos=&info->string[info->cursor];
				oldcursor-=info->cursor;
				while (*pos++=*(pos+oldcursor));
				info->length-=oldcursor;
			}
		}
		break;
	case (char)0xc8 : /* right */
		if (info->cursor<info->length)
		{
			w[0]=info->string[info->cursor];
			PFPStringWidth(gstate,w,&cw);
			info->cursor++;
			obj->curpos+=cw;
		}
		break;
	case (char)0xc9 : /* alt right */
		info->cursor=info->length;
		PFPStringWidth(gstate,info->string,&obj->curpos);
		break;
	case (char)0xca : /* ctrl right */
		{
			char *pos=&info->string[info->cursor];
			while (*pos++=*(pos+1));
			info->length--;
		}
		break;
	case (char)0xcb : /* ctrl alt right */
		info->string[info->cursor]=0;
		info->length=strlen(info->string);
		break;
	case (char)0xcc : /* shift right */
		{
			int gotspace=info->string[info->cursor]==' ';
			while ( info->cursor<info->length &&
					( (info->string[info->cursor]==' ' &&  gotspace) ||
					  (info->string[info->cursor]!=' ' && !gotspace) ) )
			{
				w[0]=info->string[info->cursor];
				PFPStringWidth(gstate,w,&cw);
				info->cursor++;
				obj->curpos+=cw;
				if (!gotspace) gotspace=info->string[info->cursor]==' ';
			}
		}
		break;
	case (char)0xce : /* shift ctrl right */
		{
			int gotspace=info->string[info->cursor]==' ';
			int endpos=info->cursor;
			while ( endpos<info->length &&
					( (info->string[endpos]==' ' &&  gotspace) ||
					  (info->string[endpos]!=' ' && !gotspace) ) )
			{
				endpos++;
				if (!gotspace) gotspace=info->string[endpos]==' ';
			}
			if (info->cursor!=endpos)
			{
				char *pos=&info->string[info->cursor];
				endpos-=info->cursor;
				while (*pos++=*(pos+endpos));
				info->length-=endpos;
			}
		}
		break;
	case (char)0xda : /* ctrl down */
		info->string[0]='\0';
		obj->curpos=info->length=info->cursor=0;
		break;
	}
	obj->cursor=info->cursor;	/* necessary for dedlines */
	return ERR_OK;
}

/* object handling routines  - common create */

Error CreateCommon(PWObject owner, PWObject *ret, PWEvent event,
					int *taglist, int this_is_an_edline)
{
	Error err;
	short buffer=bufflen;
	short llines=1; /* preset number of lines (1) */
	LinkPROforma();

	catch( PWGetType(PW_TYPE_iTEM, &border) );

	/* see if there is a PW_EDLINE_MAXLENGTH tag */
	while (*taglist)
	{
		if (*taglist==PW_EDLINE_MAXLENGTH)
		{
			buffer=(*(taglist+1));
			*taglist&=PW_TAGMASK_SKIP;
		}
		if (*taglist==PW_EDLINE_VISIBLE_LINES)
		{
			llines=(*(taglist+1));	   /* get number of lines */
			*taglist&=PW_TAGMASK_SKIP; /* this param no longer needed */
			if (llines<=0)
				llines=1;		   /* make sure there is at least 1 line */
		}

		taglist=PWSkipTag(taglist);
	}
	buffer*=llines;
	/* allocate specific object - which should be cleared (automatic) */

	catch( PWAllocate(event.region->system,sizeof(Object)+buffer,ret) );

	/* make sure the item is available */
	(*(Object **)ret)->bufflen=(buffer/llines);
	(*(Object **)ret)->xsiz=PF2xpix(xsiz);
	(*(Object **)ret)->is_edline=this_is_an_edline;
	(*(Object **)ret)->cursor=-1;

	/* preset number of lines */
	(*(Object **)ret)->lines=llines;
	(*(Object **)ret)->current_line=0;

	event.region->wait_event=PW_EVENT_HIT|PW_EVENT_DO;
	return err;
}


/* Object handling routines */
/* EDline creation */
Error Create_edline(PWObject owner, PWObject *ret, PWEvent event, int *taglist)
{
	if (always_type==2) 			/* always create dedlines */
		return CreateCommon(owner, ret, event, taglist, FALSE);
	else
		return CreateCommon(owner, ret, event, taglist, TRUE);
}

/* DEDline creation */
Error Create_dedline(PWObject owner, PWObject *ret, PWEvent event, int *taglist)
{
	if (always_type==1) 			/* always create edlines */
		return CreateCommon(owner, ret, event, taglist, TRUE);
	else
		return CreateCommon(owner, ret, event, taglist, FALSE);
}

Error Remove(PWObject object)
{
	Object *obj=(Object *)object;
	if (obj->keypress) PWRemove(obj->keypress);
	return PWRelease((PWObject)obj,(PWObject)obj);
}

Error draw_edline(Object *obj, int show)
{
	PWRegion *reg=obj->event.region;
	Gstate gstate=reg->system->gstate;
	pix hit_xsiz=reg->hit_xend-reg->hit_xorg;
	pix hit_ysiz=reg->hit_yend-reg->hit_yorg;
	pix my_height=hit_ysiz/obj->lines;
	pt ypos=short2pt(hit_ysiz-height)/2;
	short my_lines=0;
	char *t=obj->string;
	pix offset=(my_height-height)/2;	/* centre line within space for it */
	short dist, min_cur,max_cur;

	if (!gstate) return ERR_OK;

	PFPWindowSubSet(gstate,hit_xsiz,hit_ysiz,reg->hit_xorg,reg->hit_yorg);
	PFPaperColourRGB(gstate,&paper);
	PFColourRGB(gstate,&ink);
	PFPageClear(gstate);	/* clear all lines */
	PFFontLoad(gstate,font);
	PFFontScale(gstate,fontsize);

	PFPrintEscape(gstate,obj->cursor>=0?TRUE:obj->escape);

	min_cur=min(height,hit_xsiz/8);
	max_cur=hit_xsiz-min_cur;

	/* make sure cursor is visible in window */
	if ( obj->cursor>=0 &&
		 ( (dist=obj->curpos-max_cur-obj->offset)>0 ||
		   (dist=obj->curpos-min_cur-obj->offset)<0 ) )
	{
		obj->offset+=dist;
		if (obj->offset<0)
			obj->offset=0;
	}
	PFPPageOriginSet(gstate,obj->offset,0);

	while (my_lines!=obj->lines)
	{
		PFPMoveTo(gstate,0,short2pt(my_height*my_lines+offset)+baseline);
		PFPStringShow(gstate,t);
		t+=obj->bufflen;
		my_lines++;
	}

	if (obj->cursor>=0)
	{
		PFPathStroked(gstate);
		PFLineWidth(gstate,0);
		PFPMoveTo(gstate,short2pt(obj->curpos),short2pt(obj->current_line*my_height));
		PFPLineR(gstate,0,short2pt(height));
		PFPathDraw(gstate);
	}

	if (show) PFPageShow(gstate);
	PFWindowSubRestore(gstate);
	return PFPrintEscape(gstate,FALSE);
}

Error Draw(PWObject object)
{
	return draw_edline((Object *)object, FALSE);
}

Error Entry(PWObject object)
{
	/* only react for edlines */
	Object *obj=(Object *)object;
	if (obj->is_edline)
	   return border->PWRegionEntry((PWObject)obj);
	return ERR_OK;
}

Error Exit(PWObject object)
{
	/* only react for edlines */
	Object *obj=(Object *)object;
	if (obj->is_edline)
		return border->PWRegionExit((PWObject)obj);
	return ERR_OK;
}

/* handle routine */
Error Handle(PWObject object, PWWait event)
{
	Error err=ERR_OK;
	Object *obj=(Object *)object;
	PWSystem *system=obj->event.region->system;
	Gstate gstate=system->gstate;
	PWRegion *reg=obj->event.region;
	pix hit_xsiz=reg->hit_xend-reg->hit_xorg;
	pix hit_ysiz=reg->hit_yend-reg->hit_yorg;
	pix my_height=hit_ysiz/obj->lines;
	pix poffset=(my_height-height)/2; /* centre line within space for it */
	short cw;
	int line_offset;

	if (!gstate) return ERR_OK;

	PFPrintEscape(gstate,TRUE);

	switch (event)
	{
	case PW_EVENT_DO:
		if (obj->DoAction)
			return obj->DoAction((PWObject)obj);
	case PW_EVENT_HIT:
		if (obj->is_edline)
		{
			char key;
			PWObject oldcatch=system->catchkeys, curr;
			PWChange((PWObject)obj,PW_CATCH_OBJECT,obj,NULL);
			do {
				IOGet(system->window,&key);
				system->catch=key;
				curr=system->catchkeys;
				Handle(curr,PW_EVENT_CATCH);
			} while (key!='\n' || curr!=system->catchkeys);
			PWChange((PWObject)obj,PW_CATCH_OBJECT,oldcatch,NULL);
			Entry((PWObject)obj);
			/* clear keypress !! */
			{
				PointerInfo pr;
				short xp, yp;
				Error err;

				err=WINPointerRead(system->window,1,EVENT_KEYUP,&xp,&yp,&pr);
				if (err=ERR_NC && pr.event&(EVENT_KEYDOWN|EVENT_KEYPRESS))
					WINPointerRead(system->window,FOREVER,EVENT_KEYUP,&xp,&yp,&pr);
			}
		}
		else
			PWChange((PWObject)obj,PW_CATCH_OBJECT,obj,NULL);
		break;
	case PW_EVENT_CATCHSTART:
		line_offset = obj->current_line*obj->bufflen;
		if (obj->is_edline) border->PWRegionEntry((PWObject)obj);
		PFFontLoad(gstate,font);
		PFFontScale(gstate,fontsize);
		PFPStringWidth(gstate,obj->string+line_offset,&obj->curpos);
		obj->length=strlen(obj->string+line_offset);
		obj->cursor=obj->length;
		system->catch=0;
		goto catch;
	case PW_EVENT_CATCHEND:
		obj->offset=0;
		obj->cursor=-1;
		if (obj->is_edline) border->PWRegionExit((PWObject)obj);
		if (obj->DoAfter) err=obj->DoAfter((PWObject)obj);
		goto redraw;
	case PW_EVENT_CATCH:
catch:	{
			char c=system->catch;
			char w[2]={0,0};
			EdlineInfo info;

			line_offset = obj->current_line*obj->bufflen;
			info.maxlength=obj->bufflen;
			info.string=obj->string+line_offset;
			info.length=strlen(info.string);
			info.cursor=obj->cursor;
			info.key=c;
			obj->dummy=NULL;
						  
			/* Handle the keypress, allow callback mechanism to check */
			/* whether the keypress is valid */
			/* The callback routine can have three possible results */
			/* ACCEPT keypress is valid, use normal processing */
			/* NO ACCEPT do not accept */
			/* CHANGE do not use normal processing, the string and/or */
			/*		cursor position are changed by the callback routine */
			if (obj->ActionTest)
			{
				int i;
				err=obj->ActionTest((PWObject)obj,&info);
				switch(err)
				{
				case PW_EDLINE_TEST_NOACCEPT:
					return ERR_OK;
				case PW_EDLINE_TEST_ACCEPT:
					c=info.key;
					break;
				case PW_EDLINE_TEST_CHANGE:
					/* string and/or cursor position have changed */
					/* adjust internal structures */
					/* determine new value of curpos */
					obj->cursor=info.cursor;
					obj->length=info.length;
					obj->curpos=0;
					for (i=0 ; i<info.cursor ; i++)
					{
						w[0]=obj->string[i];
						PFPStringWidth(gstate,w,&cw);
						obj->curpos+=cw;
					}
					goto redraw;
				}
			}

			err=DealWithChar(obj, &info, gstate);

			if (obj->dummy)
				PWChange((PWObject)obj,PW_CATCH_OBJECT,obj->dummy,NULL);

			PFPrintEscape(gstate,FALSE);
		}
redraw: draw_edline(obj,system->activated?TRUE:FALSE);
		break;
	}
	PFPrintEscape(gstate,FALSE);
	return err;
}

static Error natural(PWObject object, EdlineInfo *info)
{
	if (isprint(info->key) && !isdigit(info->key))
		return PW_EDLINE_TEST_NOACCEPT;
	else
		return PW_EDLINE_TEST_ACCEPT;
}

static Error integer(PWObject object, EdlineInfo *info)
{
	if (info->key=='+' || info->key=='-')
	{
		if (info->string[0]=='+' || info->string[0]=='-')
		{
			info->string[0]=info->key;
			return PW_EDLINE_TEST_CHANGE;
		}
		if (info->length+1<info->maxlength)
		{
			MEMCopy(info->maxlength-1,info->string,info->string+1);
			info->string[0]=info->key;
			info->cursor++;
			info->length++;
			return PW_EDLINE_TEST_CHANGE;
		}
		return PW_EDLINE_TEST_NOACCEPT;
	}
	if (isprint(info->key) && !isdigit(info->key))
		return PW_EDLINE_TEST_NOACCEPT;
	else
		return PW_EDLINE_TEST_ACCEPT;
}

static Error number(PWObject object, EdlineInfo *info)
{
	if (info->key=='+' || info->key=='-')
	{
		if (info->string[0]=='+' || info->string[0]=='-')
		{
			info->string[0]=info->key;
			return PW_EDLINE_TEST_CHANGE;
		}
		if (info->length+1<info->maxlength)
		{
			MEMCopy(info->maxlength-1,info->string,info->string+1);
			info->string[0]=info->key;
			info->cursor++;
			info->length++;
			return PW_EDLINE_TEST_CHANGE;
		}
		return PW_EDLINE_TEST_NOACCEPT;
	}
	if (info->key=='.')
	{
		/* only accept if no decimal point yet */
		char *str;
		for (str=info->string ; *str ; str++)
			if (*str=='.') return PW_EDLINE_TEST_NOACCEPT;
		return PW_EDLINE_TEST_ACCEPT;
	}
	if (isprint(info->key) && !isdigit(info->key))
		return PW_EDLINE_TEST_NOACCEPT;
	else
		return PW_EDLINE_TEST_ACCEPT;
}

/* make this dedline the current one */
Error select(PWObject object)
{
	PWObject owner;
	PWQuery(object,PW_OBJECT_AUXILIARY,&owner);
	return PWChange(object,PW_CATCH_OBJECT,owner,NULL);
}

/* common change routine */
Error Change(PWObject object, int *taglist)
{
	Object *obj=(Object *)object;
	PWSystem *system=obj->event.region->system;
	Error err=ERR_OK;
	Gstate gstate=system->gstate;
	char unrecognised_tag=FALSE;
	int line;

	if (!obj->is_edline && !(obj->status&OSTATUS_INITIALISED))
	{
		obj->status|=OSTATUS_INITIALISED;
		PWChange((PWObject)obj,PW_CURSOR_SEPARATE,TRUE,NULL);
		if (system->catchkeys==NULL)
			PWChange((PWObject)obj,PW_CATCH_OBJECT,obj,NULL);
	}

	while (*taglist)
	{
		switch (*taglist)
		{
		case PW_EDLINE_SET_LINE:
			obj->current_line=(*(taglist+2))-1;
			newmax(obj->current_line,0);
			newmin(obj->current_line,obj->lines-1);
		case PW_EDLINE_SET:
			STRCopySafe((char *)(*(taglist+1)),obj->string+(obj->current_line*obj->bufflen),obj->bufflen);
			obj->status|=OSTATUS_REDRAW;
			system->status|=SSTATUS_REDRAW;
			break;
		case PW_EDLINE_GET_LINE:
			/* three parameters : buffer length, buffer base, line to get */
			line=(*(taglist+3))-1;
			newmax(line,0);
			newmin(line,obj->lines-1);
			goto getline;
		case PW_EDLINE_GET:
			/* two parameters : buffer length and buffer base */
			line=obj->current_line;
getline:	STRCopySafe(obj->string+(line*obj->bufflen),(char *)(*(taglist+2)),(*(taglist+1)));
			break;
		case PW_EDLINE_WIDTH:
			obj->xsiz=PF2xpix(*(taglist+1));
			system->status|=SSTATUS_WINCHANGE;
			break;
		case PW_EDLINE_WIDTH_PIX:
			obj->xsiz=(*(taglist+1));
			system->status|=SSTATUS_WINCHANGE;
			break;
		case PW_EDLINE_WIDTH_FS:
			obj->xsiz=PF2xpix(fixmul((pt)(*(taglist+1)),fontsize));
			system->status|=SSTATUS_WINCHANGE;
			break;
		case PW_EDLINE_KEYPRESS:
			/* attach keypress */
			if (!obj->keypress)
			{
				if (obj->is_edline)
					err=PWCreate(object,&obj->keypress,PW_TYPE_KEYPRESS,NULL);
				else
					err=PWCreate(object,&obj->keypress,PW_TYPE_KEYPRESS,
									PW_KEYPRESS_ACTION, select,
									PW_OBJECT_AUXILIARY, object,
									NULL);
				if (err) return err;
			}
			err= PWChange(obj->keypress,PW_KEYPRESS,(*(taglist+1)),NULL);
			break;
		case PW_EDLINE_ACTION_AFTER:
			obj->DoAfter=(Error (*)(PWObject))(*(taglist+1));
			break;
		case PW_EDLINE_ACTION_DO:
			if (obj->is_edline)
				obj->DoAction=(Error (*)(PWObject))(*(taglist+1));
			break;
		case PW_EDLINE_ACTION_TEST:
			if ((*(taglist+1))<0)
			{
				switch (*(taglist+1))
				{
				case PW_EDLINE_ACTION_TEST_NATURAL:
					obj->ActionTest=natural;
					break;
				case PW_EDLINE_ACTION_TEST_INTEGER:
					obj->ActionTest=integer;
					break;
				case PW_EDLINE_ACTION_TEST_FLOAT:
					obj->ActionTest=number;
					break;
				}
			}
			else
				obj->ActionTest=(Error (*)(PWObject, EdlineInfo *))(*(taglist+1));
			break;
		case PW_EDLINE_EDLINE_AFTER:
			obj->after=(Object *)(*(taglist+1));
			break;
		case PW_EDLINE_EDLINE_NEXT:
			obj->next=(Object *)(*(taglist+1));
			break;
		case PW_EDLINE_EDLINE_PREV:
			obj->prev=(Object *)(*(taglist+1));
			break;
		case PW_EDLINE_EDLINE_UP:
			obj->up=(Object *)(*(taglist+1));
			break;
		case PW_EDLINE_EDLINE_DOWN:
			obj->down=(Object *)(*(taglist+1));
			break;
		case PW_EDLINE_ACTIVATE:
			if (obj->is_edline)
				if (system->activated) Handle((PWObject)obj,PW_EVENT_HIT);
			else
				PWChange((PWObject)obj,PW_CATCH_OBJECT,obj,NULL);
			break;
		case PW_EDLINE_PRINT_ESCAPE:
			obj->escape=(*(taglist+1));
			break;
		case PW_EDLINE_SET_ARRAY:
			{
				int i,count=(*(taglist+1));
				int length=(*(taglist+2));
				char *str=(char *)(*(taglist+3));
				char *str2=obj->string;
				int maxcpy;
				length<obj->bufflen?maxcpy=length:maxcpy=obj->bufflen;
				if (count>obj->lines)
					count=obj->lines;
				for (i=count ; i ; i--, str+=length,str2+=obj->bufflen)
					STRCopySafe(str,str2,maxcpy);
			}
			system->status|=SSTATUS_REDRAW;
			obj->status|=OSTATUS_REDRAW;
			break;
		case PW_EDLINE_GET_ARRAY:
			{
				int i,count=(*(taglist+1));
				int length=(*(taglist+2));
				char *str=(char *)(*(taglist+3));
				char *str2=obj->string;
				int maxcpy;
				length<obj->bufflen?maxcpy=length:maxcpy=obj->bufflen;
				if (count>obj->lines)
					count=obj->lines;
				for (i=count ; i ; i--, str+=length,str2+=obj->bufflen)
					STRCopySafe(str2,str,maxcpy);
			}
			break;
		default :
			if (*taglist&(~PW_TAGMASK_SKIP))
				unrecognised_tag=TRUE;
			break;
		}
		taglist=PWSkipTag(taglist);
	}
	if (unrecognised_tag) err=ERR_IPAR;
	return err;
}


Defines defines[]={
{ "INK-COLOUR",     PW_DEFINE_COLOUR,   NULL,   &ink,       &Kink },
{ "PAPER-COLOUR",   PW_DEFINE_COLOUR,   NULL,   &paper,     &Kpaper },
{ "FONT",           PW_DEFINE_FONT,     NULL,   font,       &recalculate },
{ "FONTSIZE",       PW_DEFINE_PT,       NULL,   &fontsize,  &recalculate },
{ "MAXLENGTH",      PW_DEFINE_SHORT,    NULL,   &bufflen,   NULL },
{ "ITEMWIDTH",      PW_DEFINE_PT,       NULL,   &xsiz,      NULL },
{ "ALWAYS-TYPE",    PW_DEFINE_INT,      NULL,   &always_type,  NULL },
{ 0 }
};

/* always_type	= 0 if edline<>dedline
				= 1 if always edlines
				= 2 if always dedlines
*/


Error Define(char *name, char *value)
{
	return PWHandleDefine("EDLINE-",defines,name,value);
}

Error DetermineSize(PWObject object)
{
	PWRegion *region=object->event.region;
	Object *obj=(Object *)object;
	Gstate gstate=obj->event.region->system->gstate;

	if (!gstate) return ERR_OK;

	if (!Kink) PWQuery(object,PW_DEFAULT_FOREGROUND,&ink);
	if (!Kpaper) PWQuery(object,PW_DEFAULT_BACKGROUND,&paper);
	if (recalculate)
	{
		PBbox bbox;
		if (!fontsize) PWQuery(object,PW_DEFAULT_FONTSIZE,&fontsize);
		if (!*font) PWQuery(object,PW_DEFAULT_FONT,font);
		PFFontLoad(gstate,font);
		PFFontScale(gstate,fontsize);
		PFPFontBbox(gstate,0,&bbox);
		baseline=short2pt(bbox.yorg);
		height=bbox.ysiz;
		recalculate=FALSE;
	}
	region->tot_ysiz=short2pt(height*obj->lines);
	region->tot_xsiz=short2pt(obj->xsiz);

	if (obj->is_edline)
		border->PWDetermineSize((PWObject)obj);
	return ERR_OK;
}


PWTypeDef t_edline={
	PW_TYPEDEF_FLAG, EDTYPE, PW_TYPE_EVENT_REGION,
	&Create_edline, &Change, NULL, &Remove, &Entry, &Exit, &Handle, &Define,
	&Draw, &DetermineSize, NULL
};

PWTypeDef init={
	(int)&t_edline, DEDTYPE, PW_TYPE_EVENT_REGION,
	&Create_dedline, &Change, NULL, &Remove, &Entry, &Exit, &Handle, &Define,
	&Draw, &DetermineSize, NULL
};
/* note: they must point to the same routines, except for the creation */
/* routines else they will not behave correctly when always_type is set */
