/*
	ProWesS, PROGS Windowing System

	list select object "LSEL"
	You can give a list of items, one of which is displayed as current.

	(c) PROGS, Professional & Graphical Software
	written by Joachim Van der Auwera
	05.05.1997 started
	13.09.1997 now reduces size properly on subwin display
	21.10.1997 query also accepts all PW_MENU tags for the submenu
	08.10.1998 listselect_clear tag no longer affects the current value
*/

#include <mem.h>
#include <str.h>
#include <ctype.h>

#include "PWHandler.h"

#define catch(x) do { if (err=(x)) return err; } while (0)

/* global statusses */

#define p100 short2pt(100)
ColourRGB arrow_colour={0,0,0}; char Kcolour=FALSE;
pt xarrow=short2pt(6), yarrow=short2pt(6), width;	/* size nominal and in each direction */

typedef struct {
	PWBaseObject;
	char *value;					/* current value for item */
	Error (*Select)(PWObject object, char *item);
	PWObject loose; 				/* loose item which represents the list */
	PWObject outl, comment, menu;	/* menu to select the item */
} Object;

/* object handling routines */


Error Create(PWObject owner, PWObject *ret, PWEvent event, int *taglist)
{
	Error err;

	LinkPROforma();

	/* allocate specific object - which should be cleared (automatic) */
	catch( PWAllocate(event.region->system,sizeof(Object),ret) );

	if (!Kcolour) PWQuery(*ret,PW_DEFAULT_MIDDLEGROUND,&arrow_colour);

	/* no events to react to BUT is composite */
	event.region->composite=TRUE;

	return err;
}

Error Remove(PWObject object)
{
	Object *obj=(Object *)object;
	PWRemove(obj->outl);
	return PWRelease((PWObject)obj,(PWObject)obj);
}

Error loose_hit(PWObject object)
{
	Object *obj;
	char *first;

	PWQuery(object,PW_OBJECT_AUXILIARY,&obj);
	PWQuery(obj->menu,PW_MENU_FIRST,&first);
	if (first)
	{
		PWChange(obj->menu, PW_WINDOW_ORIGIN_RESET,
							PW_MENU_VISIBLE_LINES, 2,
							PW_MENU_ITEMWIDTH_MAX,
							PW_OBJECT_AUTOSIZE,
							NULL);
		return PWActivate(obj->outl);
	}
	else
		return ERR_OK;
}

Error loose_draw(PWObject object, Gstate gstate, pt xsiz, pt ysiz)
{
	Object *obj;
	pt xorg=xpix2PF(1);
	pt yorg=(ysiz-yarrow)/2;

	PWQuery(object,PW_OBJECT_AUXILIARY,&obj);

	if (obj->value==NULL)
	{
		char *first;
		PWQuery(obj->menu,PW_MENU_FIRST,&first);
		if (first)
			PWChange((PWObject)obj,PW_LISTSELECT_CURRENT,first,NULL);
	}

	PFPathFilled(gstate);
	PFColourRGB(gstate, &arrow_colour);
	PFMoveTo(gstate, xorg, yorg);
	PFLineR(gstate, xarrow, 0);
	PFLineTo(gstate, xorg+xarrow/2, yorg+yarrow);
	PFPathClose(gstate);
	PFPathDraw(gstate);
	return ERR_OK;
}

Error menu_select(PWObject object, char *item)
{
	Object *obj;

	PWQuery(object,PW_GLOBAL_AUXILIARY,&obj);
	PWChange((PWObject)obj,PW_LISTSELECT_CURRENT,item,NULL);
	return PWChange(object,PW_SYSTEM_BREAKDOWN,NULL);
}

Error Change(PWObject object, int *taglist)
{
	Object *obj=(Object *)object;
	PWSystem *system=obj->event.region->system;
	Error err=ERR_OK;
	char unrecognised_tag=FALSE;

	if (!(obj->status&OSTATUS_INITIALISED))
	{
		obj->status|=OSTATUS_INITIALISED;		/* init is done */

		catch( PWCreate((PWObject)obj,&obj->loose,PW_TYPE_LOOSE_ITEM,
							PW_LOOSE_CHANGE_STATUS, FALSE,
							PW_LOOSE_ACTION_HIT, loose_hit,
							PW_LOOSE_ACTION_DRAW, loose_draw,
							PW_LOOSE_CENTER_ITEM, FALSE,
							PW_LOOSE_OFFSET, xarrow+xpix2PF(2), 0,
							PW_OBJECT_AUXILIARY, obj,
							NULL) );
		catch( PWCreate(NULL,&obj->outl,PW_TYPE_OUTLINE,
							PW_OUTLINE_TITLE_TEXT, "",
							PW_OUTLINE_QUIT,
							PW_OUTLINE_QUIT_KEYPRESS, 27,
							PW_GLOBAL_AUXILIARY, obj,
							NULL) );
		catch( PWCreate(obj->outl,&obj->menu,PW_TYPE_MENU,
							PW_MENU_NONE_SELECTED,
							PW_MENU_ACTION_SELECT, menu_select,
							NULL) );
	}

	while (*taglist)
	{
		switch (*taglist)
		{
		case PW_LISTSELECT_ADD:
			PWChange(obj->menu,PW_MENU_ADD,*(taglist+1),NULL);
			break;

		case PW_LISTSELECT_ADD_COPY:
			PWChange(obj->menu,PW_MENU_ADD_COPY,*(taglist+1),NULL);
			break;

		case PW_LISTSELECT_ADD_ARRAY:
			PWChange(obj->menu,PW_MENU_ADD_ARRAY,
					 *(taglist+1),*(taglist+2),*(taglist+3),NULL);
			break;

		case PW_LISTSELECT_CLEAR:
			PWChange(obj->menu,PW_MENU_CLEAR,NULL);
			break;

		case PW_LISTSELECT_SORT_COMPARE:
			PWChange(obj->menu,PW_MENU_SORT_COMPARE,*(taglist+1),NULL);
			break;

		case PW_LISTSELECT_KEYPRESS:
			PWChange(obj->loose,PW_LOOSE_KEYPRESS,*(taglist+1),NULL);
			break;

		case PW_LISTSELECT_CURRENT:
			obj->value=(char *)(*(taglist+1));
			PWChange(obj->loose,PW_LOOSE_TEXT,obj->value,NULL);
			if (obj->Select)
				catch( obj->Select((PWObject)obj,obj->value) );
			break;

		case PW_LISTSELECT_ACTION_SELECT:
			obj->Select=(int (*)(PWObject, char *))(*(taglist+1));
			break;

		case PW_LISTSELECT_TITLE:
			PWChange(obj->outl,PW_OUTLINE_TITLE_TEXT,*(taglist+1),NULL);
			break;

		case PW_LISTSELECT_COMMENT:
			if (!obj->comment)
			{
				catch( PWCreate(obj->outl,&obj->comment,PW_TYPE_INFOSTRING,
								PW_POSITION_ABOVE, obj->menu,
								PW_SCALE_FACTOR, 0,
								NULL) );
				catch( PWCreate(obj->outl,NULL,PW_TYPE_SEPARATOR,
								PW_POSITION_ABOVE, obj->menu,
								NULL) );
			}
			PWChange(obj->comment,PW_INFOSTRING_TEXT,*(taglist+1),NULL);
			break;

		case PW_LISTSELECT_SIZE:
			PWChange(obj->loose,PW_LOOSE_SIZE,*(taglist+1),*(taglist+2),NULL);
			break;
		case PW_LISTSELECT_XSIZE:
			PWChange(obj->loose,PW_LOOSE_XSIZE,*(taglist+1),NULL);
			break;
		case PW_LISTSELECT_YSIZE:
			PWChange(obj->loose,PW_LOOSE_YSIZE,*(taglist+1),NULL);
			break;
		case PW_LISTSELECT_AUTOSIZE:
			PWChange(obj->loose,PW_LOOSE_AUTOSIZE,*(taglist+1),NULL);
			break;

		default:
			if (*taglist&(~PW_TAGMASK_SKIP))
				unrecognised_tag=TRUE;
			break;
		}
		taglist=PWSkipTag(taglist);
	}
	if (unrecognised_tag) err=ERR_IPAR;
	return err;
}

Error Query(PWObject object, int what, void *ret)
{
	Error err=ERR_OK;
	Object *obj=(Object *)object;

	switch (what)
	{
		case PW_LISTSELECT_CURRENT:
			*((char **)ret)=obj->value;
			break;
		default :
			err=PWQuery(obj->menu,what,ret);
	}
	return err;
}

void carrow(void *data)
{
	xarrow=PFxround(width);
	yarrow=PFyround(width);
}

Defines defines[]={
{ "ARROW-COLOUR", PW_DEFINE_COLOUR, NULL,   &arrow_colour, &Kcolour },
{ "ARROW-SIZE",   PW_DEFINE_PT,     carrow, &width,        NULL },
{ 0 }
};

Error Define(char *name, char *value)
{
	return PWHandleDefine("LISTSELECT-",defines,name,value);
}

PWTypeDef init={
	PW_TYPEDEF_FLAG, PW_TYPE_LISTSELECT, PW_TYPE_EVENT_REGION,
	&Create, &Change, &Query, &Remove, NULL, NULL, NULL, &Define,
	NULL, NULL, NULL
};
