/*
	ProWesS, PROGS Windowing System

	Title information object "TITL"

	(c) PROGS, Professional & Graphical Software
	written by Joachim Van der Auwera
	started 15 April 1995, based on "title-item"
	27.10.1995 no more WINCHANGE when changing title text
	22.03.1996 adjusted for internal use of pixel values
	10.04.1996 use PWHandleDefine && use ProWesS defaults when not directly set
	25.11.1996 line height now better
	07.12.1996 now use PFPShow/Width/FontBbox
*/

#include <mem.h>
#include <str.h>

#include "ProWesS.h"
#include "PWTypeDef.h"
#include "PWHandler.h"

/* global statusses */

#define p100 short2pt(100)
ColourRGB strip = { p100, p100, p100 }; char Kstrip=FALSE;
ColourRGB ink	= { 0,    0,    0    }; char Kink  =FALSE;
ColourRGB back	= { p100, 0,    0    }; char Kback =FALSE;

char font[PF_MAXFONTNAME]="";		/* name of PROforma font to use */
pt fontsize=0;						/* fontsize to use */
pt baseline=0;						/* font baseline position (in pixels) */
pix height=0;						/* height of font (at fontsize) */
char recalculate=TRUE;

/* object definition */
typedef struct {
	PWBaseObject;					/* common start for all objects */
	char *item; 					/* actual title */
	short itemwidth;				/* width of title */
} Object;

/* object handling routines */

Error Create(PWObject owner, PWObject *ret, PWEvent event, int *taglist)
{
	LinkPROforma();

	/* allocate specific object - which should be cleared (automatic) */
	return PWAllocate(event.region->system,sizeof(Object),ret);

	/* no events to react to */
}

Error Change(PWObject object, int *taglist)
{
	Object *obj=(Object *)object;
	PWSystem *system=obj->event.region->system;
	Error err=ERR_OK;
	Gstate gstate=system->gstate;
	char unrecognised_tag=FALSE;

	while (*taglist)
	{
		switch (*taglist)
		{
		case PW_TITLE_TEXT :
			obj->item=(char *)(*(taglist+1));
			obj->itemwidth=0;
			obj->status|=OSTATUS_REDRAW;
			system->status|=SSTATUS_REDRAW;
			break;
		default :
			if (*taglist&(~PW_TAGMASK_SKIP))
				unrecognised_tag=TRUE;
			break;
		}
		taglist=PWSkipTag(taglist);
	}
	if (unrecognised_tag) err=ERR_IPAR;
	return err;
}

Error Remove(PWObject object)
{
	return PWRelease(object,object);
}

Error Draw(PWObject object)
{
	Object *obj=(Object *)object;
	PWRegion *reg=obj->event.region;
	Gstate gstate=reg->system->gstate;
	pix hit_xsiz=reg->hit_xend-reg->hit_xorg;
	pix hit_ysiz=reg->hit_yend-reg->hit_yorg;
	pix stripwidth;

	if (!gstate) return ERR_OK;

	if (obj->itemwidth==0)
	{
		/* set itemwidth */
		PFFontLoad(gstate,font);
		PFFontScale(gstate,fontsize);
		PFPStringWidth(gstate,obj->item,&obj->itemwidth);
	}
	stripwidth=obj->itemwidth+1;
	if (stripwidth>hit_xsiz-6) stripwidth=hit_xsiz-6;

	PFPWindowSubSet(gstate,hit_xsiz,hit_ysiz,reg->hit_xorg,reg->hit_yorg);
	PFPaperColourRGB(gstate,&back);
	PFPageClear(gstate);
	PFWindowSubRestore(gstate);

	PFPWindowSubSet(gstate,stripwidth,height,
						  reg->hit_xorg+(hit_xsiz-stripwidth)/2,
						  reg->hit_yorg+(hit_ysiz-height)/2);
	PFPaperColourRGB(gstate,&strip);
	PFColourRGB(gstate,&ink);
	PFPageClear(gstate);
	PFFontLoad(gstate,font);
	PFFontScale(gstate,fontsize);
	PFPMoveTo(gstate,pt_one,baseline);
	PFPStringShow(gstate,obj->item);
	PFWindowSubRestore(gstate);

	return ERR_OK;
}

Defines defines[]={
{ "INK-COLOUR",     PW_DEFINE_COLOUR,   NULL,   &ink,               &Kink },
{ "PAPER-COLOUR",   PW_DEFINE_COLOUR,   NULL,   &strip,             &Kstrip },
{ "SURROUND-COLOUR",PW_DEFINE_COLOUR,   NULL,   &back,              &Kback },
{ "FONT",           PW_DEFINE_FONT,     NULL,   font,               &recalculate },
{ "FONTSIZE",       PW_DEFINE_PT,       NULL,   &fontsize,          &recalculate },
{ 0 }
};

Error Define(char *name, char *value)
{
	return PWHandleDefine("TITLE-ITEM-",defines,name,value);
}

Error DetermineSize(PWObject object)
{
	Object *obj=(Object *)object;
	PWRegion *region=obj->event.region;
	Gstate gstate=obj->event.region->system->gstate;

	region->hit_off_lft=region->hit_off_rgt=
	region->hit_off_top=region->hit_off_bot=0;

	if (!gstate) return ERR_OK;

	if (recalculate)
	{
		PBbox bbox;
		if (!fontsize) PWQuery(object,PW_DEFAULT_FONTSIZE,&fontsize);
		if (!*font) PWQuery(object,PW_DEFAULT_FONT,font);
		if (!Kstrip) PWQuery(object,PW_DEFAULT_BACKGROUND,&strip);
		if (!Kink) PWQuery(object,PW_DEFAULT_FOREGROUND,&ink);
		if (!Kback) PWQuery(object,PW_DEFAULT_MIDDLEGROUND,&back);
		PFFontLoad(gstate,font);
		PFFontScale(gstate,fontsize);
		PFPFontBbox(gstate,0,&bbox);
		baseline=short2pt(bbox.yorg);
		height=bbox.ysiz;
		recalculate=FALSE;
	}

	/* set itemwidth */
	PFFontLoad(gstate,font);
	PFFontScale(gstate,fontsize);
	PFPStringWidth(gstate,obj->item,&obj->itemwidth);

	region->tot_ysiz=short2pt(height);
	region->tot_xsiz=short2pt(obj->itemwidth+7);

	return ERR_OK;
}

PWTypeDef init={
	PW_TYPEDEF_FLAG, PW_TYPE_TITLE_ITEM, PW_TYPE_EVENT_REGION,
	&Create, &Change, NULL, &Remove, NULL, NULL, NULL, &Define,
	&Draw, &DetermineSize, NULL
};
